from dataclasses import dataclass, field
import datetime
from typing import List

from coincurve import PrivateKey, PublicKey
from flask_caching import Cache
from sqlalchemy import or_

from .battle.enums import ItemType
from .exc import InvalidMoveError, InvalidNameError
from .items import Item
from .move import (
    Buy,
    CreateNovice,
    FirstClass,
    HackAndSlash,
    LevelUp,
    Move,
    MoveDetail,
    MoveZone,
    Say,
    Sell,
    Send,
    Sleep
)
from .orm import db
from .tables import Tables
from .util import get_address


cache = Cache()


class User():
    """ It contains user's keys and avatar information. """

    private_key: PrivateKey
    public_key: PublicKey

    def __init__(self, private_key: PrivateKey, session=db.session):
        if not isinstance(private_key, PrivateKey):
            raise TypeError(
                f'private_key must be an instance of {PrivateKey.__module__}.'
                f'{PrivateKey.__qualname__}, not {private_key!r}'
            )
        self.private_key = private_key
        self.public_key: PublicKey = private_key.public_key
        self.session = session

    @property
    def address(self) -> str:
        """ It returns address of the user. """
        return get_address(self.public_key)

    def sign(self, move: Move) -> None:
        """ put signature into the given move using the user's private key. """
        if move.name is None:
            raise InvalidNameError
        move.user_public_key = self.public_key.format(compressed=True)
        move.user_address = self.address
        serialized = move.serialize(include_signature=False)
        move.signature = self.private_key.sign(
            serialized
        )
        move.id = move.hash

    @property
    def moves(self):
        """ return the user's moves. """
        return self.session.query(Move).filter(
            Move.user_public_key == self.public_key.format(compressed=True),
            Move.block != None,  # noqa: E711
        ).order_by(Move.block_id.desc())

    def move(self, new_move, tax=0, commit=True):
        """
        make a move of the user.

        :params new_move: Move object to make
        :params      tax: Tax to apply
        :params   commit: commit in this function automatically or not
        """
        new_move.tax = tax
        new_move.created_at = datetime.datetime.utcnow()
        self.sign(new_move)

        if new_move.valid:
            if commit:
                self.session.add(new_move)
                self.session.commit()
        else:
            raise InvalidMoveError

        return new_move

    def hack_and_slash(self, weapon=None, armor=None, food=None):
        details = dict()
        if weapon:
            details['weapon'] = weapon
        if armor:
            details['armor'] = armor
        if food:
            details['food'] = food
        return self.move(HackAndSlash(details=details))

    def sleep(self, spot=''):
        return self.move(Sleep())

    def send(self, item_index, amount, receiver):
        item = self.avatar.items[item_index]
        if item.amount < int(amount) or int(amount) <= 0:
            raise InvalidMoveError

        return self.move(Send(details={
            'item_index': item_index,
            'amount': amount,
            'receiver': receiver,
        }))

    def sell(self, item_index, price):
        return self.move(Sell(details={'item_index': item_index,
                                       'price': price}))

    def buy(self, move_id):
        return self.move(Buy(details={'move_id': move_id}))

    def create_novice(self, details):
        return self.move(CreateNovice(details=details))

    def first_class(self, class_):
        return self.move(FirstClass(details={'class': class_}))

    def move_zone(self, zone):
        return self.move(MoveZone(details={'zone': zone}))

    def level_up(self, new_status):
        return self.move(LevelUp(details={
            'new_status': new_status,
        }))

    def say(self, content):
        return self.move(Say(details={'content': content}))

    def avatar(self, block_id=None):
        """
        get avatar of the user.

        :params block_id: Avatar's block timing. If None, it sets last block.
        """
        if not block_id:
            from .block import Block
            block = self.session.query(Block).order_by(
                Block.id.desc()).first()
            if block:
                block_id = block.id
            else:
                block_id = 0
        return Avatar.get(self.address, block_id)


@dataclass
class Avatar:
    name: str
    user: str
    current_block: 'Block'
    class_: str  # noqa
    level: int = 1
    gold: int = 0
    exp: int = 0
    exp_max: int = 0
    hp: int = 0
    hp_max: int = 0
    strength: int = 0
    dexterity: int = 0
    intelligence: int = 0
    constitution: int = 0
    luck: int = 0
    items: List[Item] = field(default_factory=list)
    zone: str = ''
    gravatar_hash: str = 'HASH'

    @classmethod
    @cache.memoize(timeout=0)
    def get(cls, user_addr, block_id, session=db.session):
        """
        get avatar.

        :params user_addr: Avatar's user address
        :params  block_id: Avatar's block timing
        :params   session: Database session to get data
        """
        from .block import Block
        create_move = session.query(Move).filter(
            Move.user_address == user_addr,
            Move.block_id <= block_id
        ).order_by(
            Move.block_id.desc()
        ).filter(
            Move.name.like('create_%')
        ).first()
        if not create_move or block_id < create_move.block_id:
            return None
        moves = session.query(Move).filter(
            or_(Move.user_address == user_addr, Move.id.in_(
                    db.session.query(MoveDetail.move_id).filter_by(
                        key='receiver', value=user_addr)))
        ).filter(
            Move.block_id >= create_move.block_id,
            Move.block_id <= block_id
        ).order_by(Move.block_id.asc())
        avatar, result = create_move.execute(None)
        gold_added = session.query(Block).filter_by(
            creator=user_addr
        ).filter(Block.id <= block_id).count() * 8
        avatar.gold += gold_added

        for move in moves:
            if move.user_address == user_addr:
                avatar, result = move.execute(avatar)
            if (type(move) == Send and
               move.details['receiver'] == user_addr):
                avatar, result = move.receive(avatar)

        return avatar

    def get_item(self, item):
        """
        Get given item.

        :params item: item's ticker name
        """
        if item not in self.items:
            self.items[item] = 1
        else:
            self.items[item] += 1

    @property
    def profile_image_url(self):
        return f'https://www.gravatar.com/avatar/{self.gravatar_hash}?d=mm'

    @property
    def weapons(self) -> list:
        return [item for item in self.items if item.type_ == ItemType.WEAPON]

    @property
    def last_weapon(self):
        last_has = HackAndSlash.query.filter_by(
            user_address=self.user
        ).order_by(HackAndSlash.block_id.desc()).first()

        if last_has:
            weapon = last_has.details.get('weapon')
            if weapon:
                return int(weapon)
        return None

    @property
    def armors(self) -> list:
        return [item for item in self.items if item.type_ == ItemType.ARMOR]

    @property
    def last_armor(self):
        last_has = HackAndSlash.query.filter_by(
            user_address=self.user
        ).order_by(HackAndSlash.block_id.desc()).first()

        if last_has:
            armor = last_has.details.get('armor')
            if armor:
                return int(armor)
        return None

    @property
    def foods(self) -> list:
        return [item for item in self.items if item.type_ == ItemType.FOOD]

    @property
    def dead(self) -> bool:
        return self.hp <= 0

    @property
    def unlocked_zone(self) -> list:
        zone_list = []
        for (k, v) in Tables.zone.items():
            if self.level >= v.unlock_level:
                zone_list.append(k)
        return zone_list



from ..battle.enums import ItemType
from ..tables import Tables


class Item:
    subclasses = {}

    def __init_subclass__(cls):
        super().__init_subclass__()
        Item.subclasses[cls.__name__] = cls

    def __init__(self, name, amount=1):
        self.type_ = ItemType.ITEM
        self.name = name
        self.amount = amount
        self.data = Tables.items[name]


class UseItem:
    def __init__(self, name):
        super().__init__(name)

    def use(self):
        # TODO
        pass


class Equipment(Item):
    def __init__(self, name):
        super().__init__(name)
        self.is_equipped = False

    def equip(self):
        self.is_equipped = True

    def unequip(self):
        self.is_equipped = False



import dataclasses
import os

from pkg_resources import resource_string

from ..battle import WeightedList


class TableData(dict):
    def __init__(self, header, data):
        for key in header:
            data_type = type(getattr(self, key))
            setattr(self, key, data_type(data[header.index(key)]))


@dataclasses.dataclass
class ExpData(TableData):
    id: str = ''
    exp_max: int = 0

    def __init__(self, header, data):
        super().__init__(header, data)


@dataclasses.dataclass
class ItemData(TableData):
    id: str = ''
    cls: str = ''
    param_0: int = 0
    param_1: int = 0
    param_2: int = 0

    def __init__(self, header, data):
        super().__init__(header, data)


@dataclasses.dataclass
class ItemDropData(TableData):
    id: str = ''
    zone_id: str = ''
    item_id: str = ''
    weight: int = 0

    def __init__(self, header, data):
        super().__init__(header, data)


@dataclasses.dataclass
class MonsterAppearData(TableData):
    id: str = ''
    zone_id: str = ''
    monster_id: str = ''
    weight: int = 0

    def __init__(self, header, data):
        super().__init__(header, data)


@dataclasses.dataclass
class MonsterData(TableData):
    id: str = ''
    strength: int = 0
    dexterity: int = 0
    intelligence: int = 0
    constitution: int = 0
    luck: int = 0
    defense: int = 0
    skill_0: str = ''
    skill_1: str = ''
    skill_2: str = ''
    skill_3: str = ''
    reward_exp: int = 0

    def __init__(self, header, data):
        super().__init__(header, data)


@dataclasses.dataclass
class NpcData(TableData):
    id: str = ''
    level: int = 0
    class_: str = '' # noqa

    def __init__(self, header, data):
        super().__init__(header, data)


@dataclasses.dataclass
class SkillData(TableData):
    id: str = ''
    cls: str = ''
    class_: str = '' # noqa
    unlock_lv: int = 0
    cast_time: int = 0
    cooltime: int = 0
    target_count: int = 0
    power: int = 0

    def __init__(self, header, data):
        super().__init__(header, data)


@dataclasses.dataclass
class StatsData(TableData):
    id: str = ''
    strength: int = 0
    dexterity: int = 0
    intelligence: int = 0
    constitution: int = 0
    luck: int = 0
    strength_lv: int = 0
    dexterity_lv: int = 0
    intelligence_lv: int = 0
    constitution_lv: int = 0
    luck_lv: int = 0

    def __init__(self, header, data):
        super().__init__(header, data)


@dataclasses.dataclass
class ZoneData(TableData):
    id: str = ''
    unlock_level: int = 0

    def __init__(self, header, data):
        super().__init__(header, data)


class Table(dict):
    separator = '\t'

    def __init__(self, filename, datacls):
        if type(filename) == str:
            self.load(filename, datacls)
        elif type(filename) == list:
            for f in filename:
                self.load(f, datacls)

    def load(self, filename, datacls):
        text = resource_string(
            'nekoyume', os.path.join('tables', filename)
        ).decode('utf-8')
        lines = text.split('\n')
        header = lines[0].split(Table.separator)
        for line in lines[1:]:
            if line:
                data = line.split(Table.separator)
                self[data[0]] = datacls(header, data)

    def get(self, id):
        if type(id) is int:
            return super().get(str(id))
        return super().get(id)


class Tables:
    exp = Table('exp.tsv', ExpData)
    item_drop = Table('item_drop.tsv', ItemDropData)
    items = Table(['items.tsv', 'item_equips.tsv'], ItemData)
    monster_appear = Table('monster_appear.tsv', MonsterAppearData)
    monsters = Table('monsters.tsv', MonsterData)
    npc = Table('npc.tsv', NpcData)
    skills = Table(['skills.tsv', 'monster_skills.tsv'], SkillData)
    stats = Table('stats.tsv', StatsData)
    zone = Table('zone.tsv', ZoneData)

    @classmethod
    def get_exp_max(cls, level: int) -> int:
        try:
            return cls.exp[str(level)].exp_max
        except KeyError:
            return 0

    @classmethod
    def get_item_drop_list(cls, zone: str) -> WeightedList:
        drop_items = WeightedList()
        for id_ in Tables.item_drop:
            data = Tables.item_drop[id_]
            if data.zone_id == zone:
                drop_items.add(data.item_id, data.weight)
        return drop_items

    @classmethod
    def get_monster_appear_list(cls, zone: str) -> WeightedList:
        appear_monsters = WeightedList()
        for id_ in Tables.monster_appear:
            data = Tables.monster_appear[id_]
            if data.zone_id == zone:
                appear_monsters.add(data.monster_id, data.weight)
        return appear_monsters

    @classmethod
    def get_npc_list(cls) -> WeightedList:
        npc_list = WeightedList()
        for id_ in Tables.npc:
            data = Tables.npc[id_]
            npc_list.add(data.id, 1)
        return npc_list



import random

from ..items import Item
from ..items.weapons import Weapon
from ..tables import Tables
from .characters import Factory
from .components.bag import Bag
from .components.stats import Stats
from .enums import CharacterType
from .logger import Logger
from .status.item import GetItem
from .status.spawn import Spawn
from .status.zone import Zone


class Simulator:
    def __init__(self, random: random.Random, zone: str):
        self.time = 0
        self.characters = []
        self.logger = Logger()
        self.zone = zone
        self.random = random
        self.result = ''  # win, lose, finish

    def simulate(self):
        self.logger.log(Zone(id_=self.zone))

        for character in self.characters:
            self.logger.log(Spawn.from_character(character))

        while True:
            self.characters = sorted(
                self.characters,
                key=lambda c: c.get_component(Stats).calc_atk_spd(),
                reverse=True)

            for character in self.characters:
                character.tick(self)

            self.time = self.time + 1
            if self.time >= 100:
                self.result = 'finish'
                break

            is_win = True
            is_lose = True
            for character in self.characters:
                if character.type_ == CharacterType.MONSTER:
                    stats = character.get_component(Stats)
                    if not stats.is_dead():
                        is_win = False
                if character.type_ == CharacterType.PLAYER:
                    stats = character.get_component(Stats)
                    if not stats.is_dead():
                        is_lose = False

            if is_win:
                self.result = 'win'
                drop_items = Tables.get_item_drop_list(self.zone)
                for character in self.characters:
                    if character.type_ == CharacterType.PLAYER:
                        bag = character.get_component(Bag)
                        drop_item = drop_items.select(self.random)
                        if drop_item:
                            item_data = Tables.items[drop_item]
                            item = Item.subclasses[item_data.cls](drop_item)
                            # TODO item option
                            item.option = self.random.randint(1, 5)
                            bag.add(item)
                            self.logger.log(GetItem(
                                id_=character.id_,
                                item=item.name
                            ))
                break
            if is_lose:
                self.result = 'lose'
                break


class DummyBattle(Simulator):
    def __init__(self, seed):
        super().__init__(seed, 'zone_0')
        factory = Factory()
        self.characters.append(factory.create_player(
            'dummy_swordman', 'swordman', 5, [Weapon('sword_1')]))
        self.characters.append(factory.create_player(
            'dummy_mage', 'mage', 5, []))
        self.characters.append(factory.create_player(
            'dummy_mage', 'acolyte', 5, []))
        self.characters.append(factory.create_monster('slime'))
        self.characters.append(factory.create_monster('slime'))
        # self.characters.append(factory.create_monster('griffin'))



import functools
from typing import Callable

from coincurve import PublicKey
from keccak import sha3_256

from nekoyume.exc import InvalidMoveError


def get_address(public_key: PublicKey) -> str:
    """Derive an Ethereum-style address from the given public key."""
    return '0x' + sha3_256(public_key.format(False)[1:]).hexdigest()[-40:]


def ensure_block(f: Callable) -> Callable:
    @functools.wraps(f)
    def decorator(self, *args, **kwargs):
        try:
            assert self.block
        except AssertionError:
            raise InvalidMoveError
        return f(self, *args, **kwargs)
    return decorator



from ..tables import Tables
from .components import ComponentContainer
from .components.bag import Bag
from .components.behaviors import BehaviorTreeBuilder
from .components.behaviors.aggro import Aggro
from .components.behaviors.skills import Skill
from .components.stats import MonsterStats, PlayerStats, Stats
from .enums import CharacterType


class Character(ComponentContainer):
    def __init__(self):
        super().__init__()
        self.id_ = 0
        self.type_ = CharacterType.NONE
        self.class_ = ''
        self.name = 'Character'
        self.behavior = None

    def tick(self, simulator):
        if self.behavior:
            self.behavior.tick(simulator)

    def to_dict(self):
        stats = self.get_component(Stats)
        dict_ = {
            'id': self.name,
            'hp': stats.hp,
        }
        return dict_

    def to_avatar(self, avatar, hp_recover=False):
        stats = self.get_component(PlayerStats)
        avatar.level = stats.level
        avatar.strength = stats.strength
        avatar.dexterity = stats.dexterity
        avatar.intelligence = stats.intelligence
        avatar.constitution = stats.constitution
        avatar.luck = stats.luck
        avatar.exp_max = stats.exp_max
        avatar.exp = stats.exp
        avatar.hp_max = stats.calc_hp_max()
        avatar.hp = avatar.hp_max if hp_recover else stats.hp
        bag = self.get_component(Bag)
        avatar.items = bag.items

    def get_skills(self):
        stats = self.get_component(Stats)
        skills = []
        for (k, v) in Tables.skills.items():
            if v.class_ == self.class_ or v.class_ == '':
                if v.unlock_lv <= stats.level:
                    skills.append(k)
        skills.reverse()
        return skills


class Factory:
    def __init__(self):
        self.character_id = 0

    @classmethod
    def set_behavior(cls, character, skills):
        b = BehaviorTreeBuilder()
        b = b.sequence(character.name)
        b = b.condition(
            'is_dead',
            lambda b: not character.get_component(Stats).is_dead())
        b = b.selector('action')
        for skillname in skills:
            data = Tables.skills[skillname]
            skillcls = data.cls
            character.add_component(Skill.subclasses[skillcls](skillname))
            b = b.do(
                skillname,
                character.get_component(Skill.subclasses[skillcls]).tick)
        b = b.end()
        b = b.do('aggro', character.get_component(Aggro).tick)
        b = b.end()
        character.behavior = b.build()

    def create_player(self, name, class_, level, items):
        self.character_id += 1
        character = Character()
        character.id_ = self.character_id
        character.type_ = CharacterType.PLAYER
        character.class_ = class_
        character.name = name
        character.add_component(PlayerStats(class_, level, 0, 0))
        stats = character.get_component(Stats)
        stats.hp = stats.calc_hp_max()
        character.add_component(Bag(items))
        character.add_component(Aggro())
        skills = character.get_skills()
        Factory.set_behavior(character, skills)
        return character

    def create_from_avatar(self, avatar, details):
        self.character_id += 1
        character = Character()
        character.id_ = self.character_id
        character.type_ = CharacterType.PLAYER
        character.class_ = avatar.class_
        character.name = avatar.name
        character.add_component(
            PlayerStats(avatar.class_, avatar.level, avatar.hp, avatar.exp))
        bag = Bag(avatar.items)
        if 'weapon' in details:
            bag.equip(int(details['weapon']))
        if 'armor' in details:
            bag.equip(int(details['armor']))
        character.add_component(bag)
        character.add_component(Aggro())
        skills = character.get_skills()
        Factory.set_behavior(character, skills)
        return character

    def create_npc(self, id):
        self.character_id += 1
        data = Tables.npc[id]
        character = Character()
        character.id_ = self.character_id
        character.type_ = CharacterType.PLAYER
        character.class_ = data.class_
        character.name = id
        character.add_component(PlayerStats(data.class_, data.level, 0, 0))
        stats = character.get_component(Stats)
        stats.hp = stats.calc_hp_max()
        character.add_component(Bag())
        character.add_component(Aggro())
        skills = character.get_skills()
        Factory.set_behavior(character, skills)
        return character

    def create_monster(self, name):
        self.character_id += 1
        data = Tables.monsters[name]
        character = Character()
        character.id_ = self.character_id
        character.type_ = CharacterType.MONSTER
        character.class_ = name
        character.name = name
        character.add_component(MonsterStats(name))
        character.add_component(Bag())
        character.add_component(Aggro())
        skills = []
        for i in range(4):
            s = 'skill_' + str(i)
            skillcls = getattr(data, s, '')
            if skillcls:
                skills.append(skillcls)
        Factory.set_behavior(character, skills)
        return character



import math

from .....tables import Tables
from .... import WeightedList
from ....enums import AttackType, CharacterType, ItemType
from ....status.skills import Attack as StatusAttack
from ....status.skills import Casting
from ....status.skills import Heal as StatusHeal
from ....status.skills import Taunt as StatusTaunt
from ....status.stats import Dead, GetExp
from ...bag import Bag
from ...stats import MonsterStats, Stats
from .. import Behavior, BehaviorTreeStatus
from ..aggro import Aggro


class Skill(Behavior):
    subclasses = {}

    def __init_subclass__(cls):
        super().__init_subclass__()
        Skill.subclasses[cls.__name__] = cls

    def __init__(self, name):
        self.name = name
        self.data = Tables.skills[name]
        self.nexttime = self.data.cooltime
        self.is_casting = False
        self.cast_remains = 0

    def tick(self, simulator):
        pass

    def find_targets(self, simulator, filter_type):
        if not self.data.target_count:
            return []
        weightedlist = WeightedList()
        for target in simulator.characters:
            stats = target.get_component(Stats)
            aggro = target.get_component(Aggro)
            if filter_type == target.type_\
               and not stats.is_dead():
                weightedlist.add(target, aggro.value)
        if len(weightedlist) == 0:
            return []
        targets = []
        while len(targets) < self.data.target_count:
            targets.append(weightedlist.pop(simulator.random))
            if len(weightedlist) == 0:
                break
        return targets

    def calc_atk(self):
        my_stats = self.owner.get_component(Stats)
        my_bag = self.owner.get_component(Bag)
        weapon = my_bag.get_equipped(ItemType.WEAPON)
        atk = 0
        if weapon:
            atk = weapon.get_atk()
            if weapon.atk_type == AttackType.MELEE:
                atk += my_stats.calc_melee_atk()
            elif weapon.atk_type == AttackType.RANGED:
                atk += my_stats.calc_ranged_atk()
            elif weapon.atk_type == AttackType.MAGIC:
                atk += my_stats.calc_magic_atk()
        else:
            atk += my_stats.calc_melee_atk()
        atk = math.floor(atk * (self.data.power * 0.01))
        return atk

    def is_cooltime(self, simulator):
        return self.nexttime > simulator.time

    def kill(self, simulator, target):
        simulator.logger.log(Dead(id_=target.id_))
        if self.owner.type_ is CharacterType.PLAYER:
            for character in simulator.characters:
                if character.type_ is CharacterType.PLAYER:
                    stats = character.get_component(Stats)
                    monster_stats = target.get_component(MonsterStats)
                    stats.get_exp(monster_stats.data.reward_exp)
                    simulator.logger.log(GetExp(
                        exp=monster_stats.data.reward_exp))

    def casting(self, simulator):
        if self.cast_remains > 0:
            self.cast_remains -= 1
            simulator.logger.log(Casting(
                id_=self.owner.id_,
                tick_remain=self.cast_remains,
            ))
            return True
        if not self.is_casting and self.data.cast_time > 0:
            self.is_casting = True
            self.cast_remains = self.data.cast_time
            simulator.logger.log(Casting(
                id_=self.owner.id_,
                tick_remain=self.cast_remains,
            ))
            return True
        self.is_casting = False
        return False


class Attack(Skill):
    """Weapon-based attack."""
    def tick(self, simulator):
        if self.is_cooltime(simulator):
            return BehaviorTreeStatus.FAILURE
        my_stats = self.owner.get_component(Stats)
        self.nexttime += my_stats.calc_cooltime(self.data.cooltime)
        atk = self.calc_atk()
        target_type = my_stats.get_target_type()
        targets = self.find_targets(simulator, target_type)
        for index, target in enumerate(targets):
            target_stats = target.get_component(Stats)
            damaged = target_stats.damaged(atk)
            target_aggro = target.get_component(Aggro)
            target_aggro.add(self.owner.id_, 1)
            simulator.logger.log(StatusAttack(
                id_=self.owner.id_,
                time=simulator.time,
                name=self.data.id,
                value=damaged,
                target_id=target.id_,
                target_hp=target_stats.hp,
                target_remain=len(targets) - index - 1,
            ))
            if target_stats.is_dead():
                self.kill(simulator, target)
        return BehaviorTreeStatus.SUCCESS


class Spell(Skill):
    def tick(self, simulator):
        if self.is_cooltime(simulator):
            return BehaviorTreeStatus.FAILURE
        if self.casting(simulator):
            return BehaviorTreeStatus.SUCCESS
        my_aggro = self.owner.get_component(Aggro)
        my_aggro.value += 1
        my_stats = self.owner.get_component(Stats)
        self.nexttime += my_stats.calc_cooltime(self.data.cooltime)
        atk = math.floor(my_stats.calc_magic_atk() * (self.data.power * 0.01))
        target_type = my_stats.get_target_type()
        targets = self.find_targets(simulator, target_type)
        for index, target in enumerate(targets):
            target_stats = target.get_component(Stats)
            damaged = target_stats.damaged(atk)
            target_aggro = target.get_component(Aggro)
            target_aggro.add(self.owner.id_, 1)

            simulator.logger.log(StatusAttack(
                id_=self.owner.id_,
                time=simulator.time,
                name=self.data.id,
                value=damaged,
                target_id=target.id_,
                target_hp=target_stats.hp,
                target_remain=len(targets) - index - 1,
            ))
            if target_stats.is_dead():
                self.kill(simulator, target)
        return BehaviorTreeStatus.SUCCESS


class Heal(Skill):
    def tick(self, simulator):
        if self.is_cooltime(simulator):
            return BehaviorTreeStatus.FAILURE
        if self.casting(simulator):
            return BehaviorTreeStatus.SUCCESS
        my_aggro = self.owner.get_component(Aggro)
        my_aggro.value += 1
        my_stats = self.owner.get_component(Stats)
        self.nexttime += my_stats.calc_cooltime(self.data.cooltime)
        targets = self.find_targets(simulator, self.owner.type_)
        for index, target in enumerate(targets):
            target_stats = target.get_component(Stats)
            if target_stats:
                amount = math.floor(
                    my_stats.calc_magic_atk() * (self.data.power * 0.01))
                target_stats.heal(amount)
                simulator.logger.log(StatusHeal(
                    id_=self.owner.id_,
                    time=simulator.time,
                    name=self.data.id,
                    value=amount,
                    target_id=target.id_,
                    target_hp=target_stats.hp,
                    target_remain=len(targets) - index - 1,
                ))
        return BehaviorTreeStatus.SUCCESS


class Slow(Skill):
    def tick(self, simulator):
        # TODO
        return BehaviorTreeStatus.FAILURE


class Taunt(Skill):
    def tick(self, simulator):
        if self.is_cooltime(simulator):
            return BehaviorTreeStatus.FAILURE
        my_aggro = self.owner.get_component(Aggro)
        my_aggro.value += math.floor(self.data.power * 0.01)
        my_stats = self.owner.get_component(Stats)
        self.nexttime += my_stats.calc_cooltime(self.data.cooltime)
        simulator.logger.log(StatusTaunt(id_=self.owner.id_))
        return BehaviorTreeStatus.SUCCESS



import datetime
import hashlib
from typing import Callable, Optional, Type, Union

from bencode import bencode
from requests import get
from requests.exceptions import ConnectionError, Timeout
from sqlalchemy.exc import IntegrityError
from typeguard import typechecked

from . import hashcash
from .exc import (
    InvalidBlockError,
    InvalidMoveError,
    NodeUnavailable,
)
from .node import Node
from .orm import db
from .user import User


MIN_BLOCK_INTERVAL = datetime.timedelta(0, 5)
MAX_BLOCK_INTERVAL = datetime.timedelta(0, 15)
PROTOCOL_VERSION: int = 2


class Block(db.Model):
    """This object contains block information."""

    __tablename__ = 'block'
    #: block id
    id = db.Column(db.Integer, primary_key=True)
    #: protocol version
    version = db.Column(db.Integer, default=PROTOCOL_VERSION, nullable=False)
    #: current block's hash
    hash = db.Column(db.String, nullable=False, index=True, unique=True)
    #: previous block's hash
    prev_hash = db.Column(db.String, index=True)
    #: block creator's address
    creator = db.Column(db.String, nullable=False, index=True)
    #: hash of every linked move's ordered hash list
    root_hash = db.Column(db.String, nullable=False)
    #: suffix for hashcash
    suffix = db.Column(db.LargeBinary, nullable=False)
    #: difficulty of hashcash
    difficulty = db.Column(db.Integer, nullable=False)
    #: block creation datetime
    created_at = db.Column(db.DateTime, nullable=False,
                           default=datetime.datetime.utcnow)
    size_limit = 10000
    __table_args__ = (
        db.CheckConstraint(id > 0),
        db.CheckConstraint(
            db.case([
                (id == 1, prev_hash.is_(None)),
            ], else_=prev_hash.isnot(None))
        ),
        db.CheckConstraint((id == 1) | (difficulty > 0)),
    )

    @classmethod
    def deserialize(cls, serialized: dict) -> 'Block':
        return cls(
            id=serialized['id'],
            version=serialized['version'],
            creator=serialized['creator'],
            created_at=datetime.datetime.strptime(
                serialized['created_at'],
                '%Y-%m-%d %H:%M:%S.%f'
            ),
            prev_hash=serialized['prev_hash'],
            hash=serialized['hash'],
            difficulty=serialized['difficulty'],
            suffix=bytes.fromhex(serialized['suffix']),
            root_hash=serialized['root_hash'],
        )

    @property
    def valid(self) -> bool:
        """Check if this object is valid or not"""
        stamp = self.serialize() + self.suffix
        valid = (self.hash == hashlib.sha256(stamp).hexdigest())
        valid = valid and hashcash.check(stamp, self.suffix, self.difficulty)

        valid = valid and (
            len(self.serialize(True, True, True, True)) <= Block.size_limit
        )

        if self.id > 1:
            prev_block = Block.query.get(self.id - 1)
            if not prev_block:
                return False
            valid = valid and self.prev_hash == prev_block.hash

            difficulty = prev_block.difficulty
            difficulty_check_block = Block.query.get(
                max(1, self.id - 10)
            )
            avg_timedelta = (
                (self.created_at - difficulty_check_block.created_at) /
                (self.id - difficulty_check_block.id)
            )
            if avg_timedelta <= MIN_BLOCK_INTERVAL:
                valid = valid and self.difficulty == max(1, difficulty + 1)
            elif avg_timedelta > MAX_BLOCK_INTERVAL:
                valid = valid and self.difficulty == max(1, difficulty - 1)
            else:
                valid = valid and self.difficulty == difficulty
        else:
            valid = valid and self.prev_hash is None
            valid = valid and self.difficulty == 0

        valid = valid and self.root_hash == hashlib.sha256(
            ''.join(sorted((m.id for m in self.moves))).encode('utf-8')
        ).hexdigest()

        for move in self.moves:
            valid = valid and move.valid
        return valid

    def serialize(self,
                  use_bencode: bool=True,
                  include_suffix: bool=False,
                  include_moves: bool=False,
                  include_hash: bool=False):
        """
        This function serialize block.

        :param    use_bencode: check if you want to encode using bencode.
        :param include_suffix: check if you want to include suffix.
        :param  include_moves: check if you want to include linked moves.
        :param   include_hash: check if you want to include block hash.
        """
        binary = (lambda x: x) if use_bencode else bytes.hex
        serialized = dict(
            id=self.id,
            creator=self.creator,
            prev_hash=self.prev_hash,
            difficulty=self.difficulty,
            root_hash=self.root_hash,
            created_at=str(self.created_at),
            version=self.version,
        )
        if include_suffix:
            serialized['suffix'] = binary(self.suffix)

        if include_moves:
            serialized['moves'] = [m.serialize(
                use_bencode=False,
                include_signature=True,
                include_id=True,
            ) for m in self.moves]

        if include_hash:
            serialized['hash'] = self.hash

        if use_bencode:
            if self.prev_hash is None:
                del serialized['prev_hash']
            serialized = bencode(serialized)
        return serialized

    @classmethod
    def sync(cls, node: Node=None, session=db.session, echo=None) -> bool:
        """
        Sync blockchain with other node.

        :param node: sync target :class:`nekoyume.node.Node`.
        """
        from .move import Move
        if not node:
            nodes = Node.query.order_by(
                Node.last_connected_at.desc()
            ).limit(10)
        else:
            nodes = [node]

        if not nodes:
            return False

        node_last_block = None
        for n in nodes:
            try:
                response = get(
                    f"{n.url}{Node.get_blocks_endpoint}/last",
                    timeout=3
                )
                if response.status_code != 200:
                    continue
                if (not node_last_block or
                   node_last_block['id'] < response.json()['block']['id']):
                    node_last_block = response.json()['block']
                    node = n
            except (ConnectionError, KeyError, Timeout):
                continue

        last_block = session.query(Block).order_by(Block.id.desc()).first()

        if not node_last_block:
            return True

        #: If my chain is the longest one, we don't need to do anything.
        if last_block and last_block.id >= node_last_block['id']:
            return True

        if last_block:
            # TODO: Very hard to understand. fix this easily.
            try:
                branch_point = find_branch_point(node, session, last_block.id,
                                                 last_block.id)
                if branch_point == last_block.id:
                    branch_point = last_block.id
                else:
                    branch_point = find_branch_point(node, session, 0,
                                                     last_block.id)
            except NodeUnavailable:
                return
        else:
            branch_point = 0

        for block in session.query(Block).filter(Block.id > branch_point):
            for move in block.moves:
                move.block_id = None
            session.delete(block)

        # Flush the above deletions to the database.
        # If we don't flush here,
        # stale objects can be fetched when validating blocks.
        session.flush()

        from_ = branch_point + 1
        limit = 1000
        while True:
            if echo:
                echo(f'Syncing blocks...(from: {from_})')
            response = get(f"{node.url}{Node.get_blocks_endpoint}",
                           params={'from': from_,
                                   'to': from_ + limit - 1})
            if response.status_code != 200:
                break
            if len(response.json()['blocks']) == 0:
                break
            for new_block in response.json()['blocks']:
                block = Block.deserialize(new_block)

                for new_move in new_block['moves']:
                    move = session.query(Move).get(new_move['id'])
                    if not move:
                        move = Move.deserialize(new_move, block.id)
                    if not move.valid:
                        session.rollback()
                        raise InvalidMoveError
                    block.moves.append(move)
                    session.add(move)

                if not block.valid:
                    session.rollback()
                    raise InvalidBlockError
                session.add(block)
            if len(response.json()['blocks']) < 1000:
                break
            from_ += limit
            try:
                session.commit()
            except IntegrityError:
                session.rollback()
                return False

        try:
            session.commit()
        except IntegrityError:
            session.rollback()
            return False
        return True

    @classmethod
    @typechecked
    def create(
            cls: Type['Block'],
            user: User,
            moves,
            commit: bool=True,
            echo: Optional[Callable]=None,
            sleep: Union[int, float]=0.0,
    ) -> Optional['Block']:
        """ Create a block. """
        for move in moves:
            if not move.valid:
                raise InvalidMoveError(move)
        block = Block(version=PROTOCOL_VERSION)
        block.root_hash = hashlib.sha256(
            ''.join(sorted((m.id for m in moves))).encode('utf-8')
        ).hexdigest()
        block.creator = user.address
        block.created_at = datetime.datetime.utcnow()

        prev_block = user.session.query(Block).order_by(
            Block.id.desc()
        ).first()
        if prev_block:
            block.id = prev_block.id + 1
            block.prev_hash = prev_block.hash
            block.difficulty = prev_block.difficulty
            difficulty_check_block = user.session.query(Block).get(
                max(1, block.id - 10)
            )
            avg_timedelta = (
                (block.created_at - difficulty_check_block.created_at) /
                (block.id - difficulty_check_block.id)
            )
            if echo:
                echo(
                    f'avg: {avg_timedelta}, difficulty: {block.difficulty}'
                )
            if avg_timedelta <= MIN_BLOCK_INTERVAL:
                block.difficulty = max(1, block.difficulty + 1)
            elif avg_timedelta > MAX_BLOCK_INTERVAL:
                block.difficulty = max(1, block.difficulty - 1)
        else:
            #: Genesis block
            block.id = 1
            block.prev_hash = None
            block.difficulty = 0
            sleep = 0

        block.suffix = hashcash._mint(block.serialize(), bits=block.difficulty,
                                      sleep=sleep)
        if user.session.query(Block).get(block.id):
            return None
        block.hash = hashlib.sha256(
            block.serialize() + block.suffix
        ).hexdigest()

        for move in moves:
            move.block = block

        if not block.valid:
            raise InvalidBlockError

        if commit:
            try:
                user.session.add(block)
                user.session.commit()
            except IntegrityError:
                return None

        return block


def find_branch_point(
        node: Node, session, value: int, high: int
) -> int:
    if value > high:
        return 0
    mid = int((value + high) / 2)
    response = get(f"{node.url}{Node.get_blocks_endpoint}/{mid}")
    if response.status_code != 200:
        raise NodeUnavailable
    block = session.query(Block).get(mid)
    if (
            response.json()['block'] and block and
            block.hash == response.json()['block']['hash']
    ):
        if value == mid:
            return value
        return find_branch_point(node, session, mid, high)
    else:
        return find_branch_point(node, session, value, mid - 1)



"""
Move
====

`move.py` contains every relations regarding nekoyume blockchain and
game moves.
"""
import datetime
import hashlib
import os
import random
import re

from bencode import bencode
from coincurve import PublicKey
from requests import get
from requests.exceptions import ConnectionError, Timeout
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm.collections import attribute_mapped_collection

from .battle.characters import Factory as CharacterFactory
from .battle.simul import Simulator
from .exc import InvalidMoveError, OutOfRandomError
from .orm import db
from .tables import Tables
from .util import ensure_block, get_address


NUM_HACK_AND_SLASH_MONSTERS: int = 3


def get_my_public_url():
    if 'PUBLIC_URL' in os.environ:
        return os.environ['PUBLIC_URL']
    try:
        if os.environ.get('PORT', '80') != '80':
            port = ':' + os.environ.get('PORT', '80')
        else:
            port = ''
        ip = get('http://ip.42.pl/raw').text
        has_public_address = get(
            f'http://{ip}{port}/ping'
        ).text == 'pong'
    except (ConnectionError, Timeout):
        return None
    if has_public_address:
        return f'http://{ip}{port}'
    else:
        return None


class Move(db.Model):
    """This object contain general move information."""
    __tablename__ = 'move'
    #: move's hash
    id = db.Column(db.String, primary_key=True)
    #: move's block id. if the move isn't confirmed yet, this will be null
    block_id = db.Column(db.Integer, db.ForeignKey('block.id'),
                         nullable=True, index=True)
    #: move's block
    block = db.relationship('Block', uselist=False, backref='moves')
    #: 33 bytes long form (i.e., compressed from) of user's public key.
    user_public_key = db.Column(db.LargeBinary, nullable=False, index=True)
    #: user's address ("0x"-prefixed 40 hexdecimal string; total 42 chars)
    user_address = db.Column(db.String, nullable=False, index=True)
    #: move's signature (71 bytes)
    signature = db.Column(db.LargeBinary, nullable=False)
    #: move name
    name = db.Column(db.String, nullable=False, index=True)
    #: move details. it contains parameters of move
    details = association_proxy(
        'move_details', 'value',
        creator=lambda k, v: MoveDetail(key=k, value=v)
    )
    #: move tax (not implemented yet)
    tax = db.Column(db.BigInteger, default=0, nullable=False)
    #: move creation datetime.
    created_at = db.Column(db.DateTime, nullable=False,
                           default=datetime.datetime.utcnow)

    __table_args__ = (
        db.CheckConstraint(
            db.func.lower(user_address).like('0x%') &
            (db.func.length(user_address) == 42)
            # TODO: it should has proper test if 40-hex string
        ),
        db.CheckConstraint(db.func.length(user_public_key) == 33),
        db.CheckConstraint(
            (db.func.length(signature) >= 68) &
            (db.func.length(signature) <= 71)
        ),
    )
    __mapper_args__ = {
        'polymorphic_identity': 'move',
        'polymorphic_on': name,
    }

    @classmethod
    def deserialize(cls, serialized: dict, block_id=None) -> 'Move':
        if block_id is None and serialized.get('block'):
            block_id = serialized['block'].get('id')
        return cls(
            id=serialized['id'],
            user_address=serialized['user_address'],
            name=serialized['name'],
            user_public_key=bytes.fromhex(serialized['user_public_key']),
            signature=bytes.fromhex(serialized['signature']),
            tax=serialized['tax'],
            details=serialized['details'],
            created_at=datetime.datetime.strptime(
                serialized['created_at'],
                '%Y-%m-%d %H:%M:%S.%f'),
            block_id=block_id,
        )

    @property
    def valid(self):
        """Check if this object is valid or not"""
        if not self.signature:
            return False

        assert isinstance(self.signature, bytes)
        assert 68 <= len(self.signature) <= 71
        assert isinstance(self.user_public_key, bytes)
        assert len(self.user_public_key) == 33
        assert isinstance(self.user_address, str)
        assert re.match(r'^(?:0[xX])?[0-9a-fA-F]{40}$', self.user_address)
        public_key = PublicKey(self.user_public_key)
        verified = public_key.verify(
            self.signature,
            self.serialize(include_signature=False),
        )
        if not verified:
            return False

        if get_address(public_key) != self.user_address:
            return False

        return self.id == self.hash

    @property
    def confirmed(self):
        """Check if this object is confirmed or not"""
        return self.block and self.block.hash is not None

    def serialize(self,
                  use_bencode=True,
                  include_signature=False,
                  include_id=False,
                  include_block=False):
        """
        This function serialize block.

        :param       use_bencode: check if you want to encode using bencode.
        :param include_signature: check if you want to include signature.
        :param        include_id: check if you want to include linked moves.
        :param     include_block: check if you want to include block.
        """
        binary = (lambda x: x) if use_bencode else bytes.hex
        serialized = dict(
            user_address=self.user_address,
            name=self.name,
            details={k: str(v) for k, v in dict(self.details).items()},
            tax=self.tax,
            created_at=str(self.created_at),
        )
        if include_signature:
            serialized.update(
                signature=binary(self.signature),
                user_public_key=binary(self.user_public_key),
            )
        if include_id:
            serialized['id'] = self.id
        if include_block:
            if self.block:
                serialized['block'] = self.block.serialize(False)
            else:
                serialized['block'] = None
        if use_bencode:
            serialized = bencode(serialized)
        return serialized

    @property
    def hash(self) -> str:
        """ Get move hash """
        return hashlib.sha256(
            self.serialize(include_signature=True)
        ).hexdigest()

    def get_randoms(self) -> list:
        """ get random numbers by :doc:`Hash random <white_paper>` """
        if not (self.block and self.block.hash and self.id):
            return []
        result = [ord(a) ^ ord(b) for a, b in zip(self.block.hash, self.id)]
        result = result[self.block.difficulty // 4:]
        return result

    def make_random_generator(self) -> random.Random:
        if self.block and self.block.hash and self.id:
            bh = bytes.fromhex(self.block.hash)
            mi = bytes.fromhex(self.id)
            seed = bytes(a ^ b for a, b in zip(bh, mi))
        else:
            seed = 0
        return random.Random(seed)

    def roll(self, randoms: list, dice: str, combine=True):
        """
        Roll dices based on given randoms

            >>> from nekoyume.move import Move
            >>> move = Move()
            >>> move.roll([1, 7, 3], '2d6')
            6

        :params randoms: random numbers from
                         :func:`nekoyume.move.Move.get_randoms`
        :params    dice: dice to roll (e.g. 2d6)
        :params combine: return combined result or not if rolling it multiple.
        """
        result = []
        if dice.find('+') > 0:
            dice, plus = dice.split('+')
            plus = int(plus)
        else:
            plus = 0
        cnt, dice_type = (int(i) for i in dice.split('d'))
        for i in range(cnt):
            try:
                result.append(randoms.pop() % dice_type + 1)
            except IndexError:
                raise OutOfRandomError
        if combine:
            return sum(result) + plus
        else:
            return result

    @ensure_block
    def execute(self):
        raise NotImplementedError()


class MoveDetail(db.Model):
    """ This object contains move's key/value information. """

    #: move id
    move_id = db.Column(db.String,  db.ForeignKey('move.id'),
                        nullable=True, primary_key=True)
    move = db.relationship(Move, backref=db.backref(
        'move_details',
        collection_class=attribute_mapped_collection("key"),
        cascade="all, delete-orphan"
    ))
    #: MoveDetail's key
    key = db.Column(db.String, nullable=False, primary_key=True)
    #: MoveDetail's value
    value = db.Column(db.String, nullable=False, index=True)


class HackAndSlash(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'hack_and_slash',
    }

    @ensure_block
    def execute(self, avatar=None):
        if not avatar:
            from .user import Avatar
            avatar = Avatar.get(self.user_address, self.block_id - 1)
        if avatar.dead:
            raise InvalidMoveError
        # TODO Load other users avatar
        rand = self.make_random_generator()
        simul = Simulator(rand, avatar.zone)

        factory = CharacterFactory()
        my_character = factory.create_from_avatar(
            avatar, self.details)
        simul.characters.append(my_character)
        appear_monsters = Tables.get_monster_appear_list(avatar.zone)
        for i in range(NUM_HACK_AND_SLASH_MONSTERS):
            simul.characters.append(
                factory.create_monster(appear_monsters.select(rand)))
        simul.simulate()

        my_character.to_avatar(avatar)

        return (avatar, dict(
                    type='hack_and_slash',
                    result=simul.result,
                    battle_logger=simul.logger,
                ))


class Sleep(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'sleep',
    }

    @ensure_block
    def execute(self, avatar=None):
        if not avatar:
            from .user import Avatar
            avatar = Avatar.get(self.user_address, self.block_id - 1)
        avatar.hp = avatar.hp_max
        return avatar, dict(
            type='sleep',
            result='success',
        )


class CreateNovice(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'create_novice',
    }

    @ensure_block
    def execute(self, avatar=None):
        from .user import Avatar
        gold = getattr(avatar, 'gold', 0)

        name = self.details.get('name', '')[:10] + '#' + self.user_address[:6]
        avatar = Avatar(
            name=name,
            user=self.user_address,
            current_block=self.block,
            gold=gold,
            class_='novice',
            level=1,
            zone=list(Tables.zone.keys())[0],
            gravatar_hash=self.details.get('gravatar_hash', 'HASH'),
        )

        factory = CharacterFactory()
        character = factory.create_from_avatar(avatar, self.details)
        character.to_avatar(avatar, hp_recover=True)

        return (avatar, dict(
            type='create_novice',
            result='success',
        ))


class FirstClass(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'first_class',
    }

    @ensure_block
    def execute(self, avatar=None):
        if not avatar:
            from .user import Avatar
            avatar = Avatar.get(self.user_address, self.block_id - 1)
        if avatar.class_ != 'novice':
            return avatar, dict(
                type='first_class',
                result='failed',
                message="Already change class.",
            )
        avatar.class_ = self.details['class']
        return avatar, dict(
            type='first_class',
            result='success',
        )


class MoveZone(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'move_zone',
    }

    @ensure_block
    def execute(self, avatar=None):
        if not avatar:
            from .user import Avatar
            avatar = Avatar.get(self.user_address, self.block_id - 1)
        zone = self.details['zone']
        if zone not in Tables.zone:
            return avatar, dict(
                type='move_zone',
                result='failed',
                message="Invalid zone.",
            )
        avatar.zone = zone
        return avatar, dict(
            type='move_zone',
            result='success',
        )


class LevelUp(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'level_up',
    }

    @ensure_block
    def execute(self, avatar=None):
        if not avatar:
            from .user import Avatar
            avatar = Avatar.get(self.user_address, self.block_id - 1)
        exp_max = Tables.get_exp_max(avatar.level)
        if exp_max == 0:
            return avatar, dict(
                type='level_up',
                result='failed',
                message="Max level.",
            )
        if avatar.exp < exp_max:
            return avatar, dict(
                type='level_up',
                result='failed',
                message="You don't have enough exp.",
            )
        avatar.exp -= exp_max
        avatar.level += 1
        setattr(avatar, self.details['new_status'],
                getattr(avatar, self.details['new_status']) + 1)
        return avatar, dict(
            type='level_up',
            result='success',
        )


class Say(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'say',
    }

    @ensure_block
    def execute(self, avatar=None):
        if not avatar:
            from .user import Avatar
            avatar = Avatar.get(self.user_address, self.block_id - 1)

        return avatar, dict(
            type='say',
            message=self.details['content'],
        )


class Send(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'send',
    }

    @ensure_block
    def execute(self, avatar=None):
        if not avatar:
            from .user import Avatar
            avatar = Avatar.get(self.user_address, self.block_id - 1)

        if int(self.details['amount']) <= 0:
            return avatar, dict(
                type='send',
                result='fail',
                message="You can't send items with a negative or zero amount."
            )

        if (self.details['item_index'] not in avatar.items or
           avatar.items[self.details['item_index']]
           - int(self.details['amount']) < 0):
            return avatar, dict(
                type='send',
                result='fail',
                message="You don't have enough items to send."
            )

        avatar.items[self.details['item_index']] -= int(self.details['amount'])
        return avatar, dict(
            type='send',
            result='success',
        )

    def receive(self, receiver=None):
        if not receiver:
            from .user import Avatar
            receiver = Avatar.get(self.details['receiver'], self.block_id - 1)

        for i in range(int(self.details['amount'])):
            receiver.get_item(self.details['item_name'])

        return receiver, dict(
            type='receive',
            result='success',
        )


class Sell(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'sell',
    }


class Buy(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'buy',
    }



#!/usr/bin/env python3.6
"""
Generalized Hashcash
====================

The function :func:`check()` will validate *generalized Hashcash* tokens
generically.

A *generalized Hashcash* is implemented in the :func:`_mint()` function.
The generalized form simply finds a suffix that creates zero bits in the
hash of the string concatenating *challenge* and *suffix* without specifying
any particular fields or delimiters in *challenge*.  E.g., you might get:

    >>> from nekoyume.hashcash import _mint
    >>> _mint(b'foo', bits=16)
    b'-g'
    >>> from hashlib import sha256
    >>> sha256(b'foo-g').hexdigest()
    '000090605c0a82a7aeac7bd99cb61002f16ced96e649edd58b8baaa3c747304a'

Notice that :func:`_mint()` behaves deterministically, finding the same suffix
every time it is passed the same arguments.
"""
import hashlib
import math
import sys
import time
import typing


def _mint(challenge: bytes, bits: int,
          sleep: typing.Union[int, float]=0) -> bytes:
    """Answer a generalized Hashcash_ challenge.

    This function accepts a generalized prefix *challenge*,
    and returns only a suffix that produces the requested SHA leading zeros.

    .. _Hashcash: https://en.wikipedia.org/wiki/Hashcash

    """
    if not isinstance(challenge, bytes):
        raise TypeError(
            f'challenge must be an instance of bytes, not {challenge}'
        )
    # These function aliases purpose to prevent global lookup which is way
    # slower than local lookup in Python VM.
    log2 = math.log2
    sha256 = hashlib.sha256
    byteorder = sys.byteorder

    counter = 1
    time.sleep(sleep)
    while 1:
        answer_bytes_length = 1 + int(log2(counter) // 8)
        answer = counter.to_bytes(answer_bytes_length, byteorder)
        digest = sha256(challenge + answer).digest()
        if has_leading_zero_bits(digest, bits):
            return answer
        counter += 1


def check(stamp, resource=None, bits=None,
          check_expiration=None, ds_callback=None):
    if type(bits) is not int:
        return True
    elif resource is not None and not stamp.endswith(resource):
        return False
    else:
        return has_leading_zero_bits(hashlib.sha256(stamp).digest(), bits)


def has_leading_zero_bits(digest: bytes, bits: int) -> bool:
    leading_bytes = bits // 8
    trailing_bits = bits % 8
    if not digest.startswith(b'\0' * leading_bytes):
        return False
    if trailing_bits:
        if len(digest) < leading_bytes:
            return False
        mask = 0b1111_1111 << (8 - trailing_bits) & 0xff
        return not (mask & digest[leading_bytes])
    return True



from ..components import Component
from ..enums import ItemType


class Bag(Component):
    def __init__(self, items=None):
        if items is None:
            items = []
        self.items = items

    def get(self, index):
        try:
            return self.items[index]
        except IndexError:
            return None

    def find(self, name):
        for item in self.items:
            if item.name == name:
                return item
        return None

    def add(self, item):
        if item.type_ == ItemType.ITEM:
            my_item = self.find(item.name)
            if my_item is None:
                self.items.append(item)
            else:
                my_item.amount += item.amount
        else:
            self.items.append(item)

    def get_equipped(self, type_):
        for item in self.items:
            if item.type_ == type_ and item.is_equipped:
                return item

    def equip(self, item_index):
        item = self.get(item_index)
        if item:
            equipped = self.get_equipped(item.type_)
            if equipped:
                equipped.unequip()
            item.equip()

    @property
    def armor_name(self):
        return getattr(self.get_equipped(ItemType.ARMOR), 'name', '')

    @property
    def head_name(self):
        return getattr(self.get_equipped(ItemType.HEAD), 'name', '')

    @property
    def weapon_name(self):
        return getattr(self.get_equipped(ItemType.WEAPON), 'name', '')



from dataclasses import dataclass

from .base import Status


@dataclass
class Dead(Status):
    type: str = 'dead'
    id_: str = ''


@dataclass
class GetExp(Status):
    type: str = 'get_exp'
    exp: int = 0



import random
import sys

from nekoyume.battle.simul import DummyBattle


def main():
    seed = 1
    if len(sys.argv) >= 2:
        seed = int(sys.argv[1])
    simulator = DummyBattle(random.Random(seed))
    simulator.logger.print = True
    simulator.simulate()


if __name__ == '__main__':
    main()



import math

from ...tables import Tables
from ..enums import CharacterType
from . import Component


class Stats(Component):
    def __init__(self):
        self.is_damaged = False

    def get_target_type(self):
        player = self.owner.type_ == CharacterType.PLAYER
        return CharacterType.MONSTER if player else CharacterType.PLAYER

    def calc_melee_atk(self):
        return self.strength + math.floor(self.dexterity * 0.2)

    def calc_ranged_atk(self):
        return self.dexterity + math.floor(self.strength * 0.2)

    def calc_magic_atk(self):
        return self.intelligence

    def calc_atk_spd(self):
        return self.dexterity

    def calc_cooltime(self, cooltime):
        return cooltime

    def calc_defense(self):
        return self.data.get('defense', 0)

    def calc_hp_max(self):
        return self.constitution * 10

    def damaged(self, damage):
        damage -= self.calc_defense()
        self.hp -= damage
        self.is_damaged = True
        return damage

    def check_damaged(self):
        damaged = self.is_damaged
        self.is_damaged = False
        return damaged

    def heal(self, amount):
        self.hp = min(self.hp + amount, self.calc_hp_max())

    def is_dead(self):
        return self.hp <= 0


class PlayerStats(Stats):
    def __init__(self, class_, level, hp, exp):
        self.data = Tables.stats.get(class_)
        super().__init__()
        self.level = level
        self.hp = hp
        self.exp = exp
        self.exp_max = Tables.get_exp_max(level)

    def get_exp(self, exp):
        self.exp += exp
        while self.exp >= self.exp_max:
            self.exp -= self.exp_max
            self.level += 1
        self.exp_max = Tables.get_exp_max(self.level)

    @property
    def strength(self):
        return self.data.strength + \
            self.data.strength_lv * (self.level - 1)

    @property
    def dexterity(self):
        return self.data.dexterity + \
            self.data.dexterity_lv * (self.level - 1)

    @property
    def intelligence(self):
        return self.data.intelligence + \
            self.data.intelligence_lv * (self.level - 1)

    @property
    def constitution(self):
        return self.data.constitution + \
            self.data.constitution_lv * (self.level - 1)

    @property
    def luck(self):
        return self.data.luck + \
            self.data.luck_lv * (self.level - 1)


class MonsterStats(Stats):
    def __init__(self, name):
        self.data = Tables.monsters.get(name)
        super().__init__()
        self.name = name
        self.hp = self.calc_hp_max()

    @property
    def strength(self):
        return self.data.strength

    @property
    def dexterity(self):
        return self.data.dexterity

    @property
    def intelligence(self):
        return self.data.intelligence

    @property
    def constitution(self):
        return self.data.constitution

    @property
    def luck(self):
        return self.data.luck



from dataclasses import dataclass

from .base import Status


@dataclass
class Skill(Status):
    type: str = ''
    name: str = ''


@dataclass
class Attack(Skill):
    type: str = 'attack'
    id_: str = ''
    value: int = 0
    target_id: str = ''
    target_hp: int = 0
    target_remain: int = 0


@dataclass
class Casting(Skill):
    type: str = 'casting'
    id_: str = ''
    tick_remain: int = 0


@dataclass
class Delaying(Skill):
    type: str = 'delaying'
    id_: str = ''
    tick_remain: int = 0


@dataclass
class Taunt(Skill):
    type: str = 'taunt'
    id_: str = ''


@dataclass
class Heal(Skill):
    type: str = 'heal'
    id_: str = ''
    value: int = 0
    target_id: str = ''
    target_hp: int = 0
    target_remain: int = 0



from dataclasses import dataclass

from ..characters import Character
from ..components.bag import Bag
from ..components.stats import Stats
from .base import Status


@dataclass
class Spawn(Status):
    type: str = 'spawn'
    id_: str = ''
    class_: str = '' # noqa
    character_type: int = 0
    name: str = ''
    level: int = 0
    hp: int = 0
    hp_max: int = 0
    armor: str = ''
    head: str = ''
    weapon: str = ''

    @classmethod
    def from_character(cls, character: Character):
        bag = character.get_component(Bag)
        stats = character.get_component(Stats)
        return Spawn(
            id_=character.id_,
            class_=character.class_,
            character_type=character.type_,
            name=character.name,
            level=getattr(stats, 'level', 0),
            hp=stats.hp,
            hp_max=stats.calc_hp_max(),
            armor=bag.armor_name,
            head=bag.head_name,
            weapon=bag.weapon_name,
        )



import os

from flask import Flask
from sentry_sdk import init as sentry_init
from sentry_sdk.integrations.flask import FlaskIntegration

from .api import api
from .game import babel, game
from .orm import db
from .tasks import celery
from .user import cache


def make_celery(app):
    celery.name = app.import_name
    celery.conf.update(
        result_backend=app.config['CELERY_RESULT_BACKEND'],
        broker_url=app.config['CELERY_BROKER_URL'])
    celery.conf.update(app.config)
    TaskBase = celery.Task

    class ContextTask(TaskBase):
        abstract = True

        def __call__(self, *args, **kwargs):
            with app.app_context():
                return TaskBase.__call__(self, *args, **kwargs)

    celery.Task = ContextTask
    return celery


def create_app():
    if os.environ.get('SENTRY_DSN'):
            sentry_init(
                dsn=os.environ['SENTRY_DSN'],
                integrations=[FlaskIntegration()]
            )
    app = Flask(__name__)
    app.config['SQLALCHEMY_DATABASE_URI'] = os.environ.get(
        'DATABASE_URL', 'sqlite:///yume.db')
    app.config['TEMPLATES_AUTO_RELOAD'] = True
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

    try:
        with open('.secret_key', 'rb') as f:
            app.secret_key = f.read()
    except FileNotFoundError:
        app.secret_key = os.urandom(64)
        f = open('.secret_key', 'wb')
        f.write(app.secret_key)
        f.close()

    app.register_blueprint(api)
    app.register_blueprint(game)

    db.app = app
    db.init_app(app)

    babel.app = app
    babel.init_app(app)

    app.config.update(
        CELERY_BROKER_URL=os.environ.get(
            'CELERY_BROKER_URL', 'sqla+sqlite:///yume_broker.db'),
        CELERY_RESULT_BACKEND=os.environ.get(
            'CELERY_RESULT_BACKEND', 'db+sqlite:///yume_reuslt.db'))
    return app


app = create_app()
cel = make_celery(app)

cache_type = os.environ.get('CACHE_TYPE', 'filesystem')
if cache_type == 'redis':
    cache_config = {
        'CACHE_TYPE': cache_type,
        'CACHE_REDIS_URL': os.environ.get(
            'REDIS_URL', 'redis://localhost:6379'
        ),
    }
else:
    cache_config = {
        'CACHE_TYPE': cache_type,
        'CACHE_DIR': os.environ.get('CACHE_DIR', '.yumecache'),
    }

cache.init_app(app, cache_config)


def run():
    from gunicorn.app import wsgiapp
    wsgiapp.run()



import datetime

from flask import Blueprint, jsonify, request
from requests import get
from requests.exceptions import ConnectionError
from sqlalchemy.exc import IntegrityError

from .block import Block
from .broadcast import (
    POST_BLOCK_ENDPOINT,
    POST_MOVE_ENDPOINT,
    POST_NODE_ENDPOINT,
)
from .move import Move, get_my_public_url
from .node import Node
from .orm import db
from .tasks import block_broadcast, move_broadcast


api = Blueprint('api', __name__, template_folder='templates')


@api.route('/ping')
def get_pong():
    return 'pong'


@api.route('/public_url')
def get_public_url():
    return jsonify(
        url=get_my_public_url()
    )


@api.route(Node.get_nodes_endpoint, methods=['GET'])
def get_nodes():
    nodes = Node.query.filter(
        Node.last_connected_at >= datetime.datetime.utcnow() -
        datetime.timedelta(minutes=60 * 3)
    ).order_by(Node.last_connected_at.desc()).limit(2500).all()

    nodes = [n.url for n in nodes]

    public_url = get_my_public_url()
    if public_url and public_url not in nodes:
        nodes.append(public_url)

    return jsonify(nodes=nodes)


@api.route(POST_NODE_ENDPOINT, methods=['POST'])
def post_node():
    if 'url' in request.values:
        url = request.values['url']
    else:
        payload = request.get_json()
        try:
            url = payload['url']
        except (KeyError, TypeError):
            return jsonify(
                result='failed',
                message='Invalid parameter.'
            ), 400
    node = Node.query.get(url)
    if not node:
        node = Node(url=url)
        db.session.add(node)
    try:
        response = get(f'{node.url}/ping')
    except ConnectionError:
        db.session.rollback()
        return jsonify(
            result='failed',
            message=f'Connection to node {node.url} was failed.'
        ), 403
    if response.status_code == 200:
        node.last_connected_at = datetime.datetime.utcnow()
        db.session.commit()
        return jsonify(result='success')
    else:
        db.session.rollback()
        return jsonify(
            result='failed',
            message=f'Connection to node {node.url} was failed.'
        ), 403


@api.route(Node.get_blocks_endpoint, methods=['GET'])
def get_blocks():
    last_block = Block.query.order_by(Block.id.desc()).first()
    from_ = request.values.get('from', 1, type=int)
    to = request.values.get(
        'to',
        last_block.id if last_block else 0,
        type=int)
    blocks = Block.query.filter(
        Block.id >= from_,
        Block.id <= to
    ).order_by(Block.id.asc())
    return jsonify(blocks=[b.serialize(use_bencode=False,
                                       include_suffix=True,
                                       include_moves=True,
                                       include_hash=True)
                           for b in blocks])


@api.route('/blocks/<string:block_hash>')
def get_block_by_hash(block_hash):
    block = Block.query.filter_by(hash=block_hash).first()
    if block:
        block = block.serialize(use_bencode=False,
                                include_suffix=True,
                                include_moves=True,
                                include_hash=True)
    return jsonify(block=block)


@api.route('/blocks/<int:block_id>')
def get_block_by_id(block_id):
    block = Block.query.get(block_id)
    if block:
        block = block.serialize(use_bencode=False,
                                include_suffix=True,
                                include_moves=True,
                                include_hash=True)
    return jsonify(block=block)


@api.route('/blocks/last')
def get_last_block():
    block = Block.query.order_by(Block.id.desc()).first()
    if block:
        block = block.serialize(use_bencode=False,
                                include_suffix=True,
                                include_moves=True,
                                include_hash=True)
    return jsonify(block=block)


@api.route('/moves/<string:move_id>')
def get_moves(move_id):
    move = Move.query.get(move_id)
    if move:
        move = move.serialize(False, True, True, True)
    return jsonify(move=move)


@api.route(POST_BLOCK_ENDPOINT, methods=['POST'])
def post_block():
    new_block = request.get_json()
    last_block = Block.query.order_by(Block.id.desc()).first()

    if not new_block:
        return jsonify(result='failed',
                       message="empty block."), 400

    if not last_block and new_block['id'] != 1:
        Block.sync(
            Node.query.order_by(
                Node.last_connected_at.desc()).first())
        return jsonify(result='failed',
                       message="new block isn't our next block."), 403

    if (new_block['id'] > 1 and
       (new_block['id'] != last_block.id + 1 or
       new_block['prev_hash'] != last_block.hash)):
        if new_block['id'] > last_block.id + 1:
            Block.sync(
                Node.query.order_by(
                    Node.last_connected_at.desc()).first())
        return jsonify(result='failed',
                       block_id=last_block.id,
                       message="new block isn't our next block."), 403

    block = Block.deserialize(new_block)

    for new_move in new_block['moves']:
        move = Move.query.get(new_move['id'])
        if not move:
            Move.deserialize(new_move, block.id)
        if not move.valid:
            return jsonify(result='failed',
                           message=f"move {move.id} isn't valid."), 400
        block.moves.append(move)
    if not block.valid:
        return jsonify(result='failed',
                       message="new block isn't valid."), 400

    db.session.add(block)
    try:
        db.session.commit()
    except IntegrityError:
        return jsonify(result='failed',
                       message="This node already has this block."), 400
    sent_node = Node()
    if 'sent_node' in new_block:
        sent_node.url = new_block['sent_node']
    block_broadcast.delay(
        block.id,
        sent_node_url=sent_node.url,
        my_node_url=f'{request.scheme}://{request.host}'
    )
    return jsonify(result='success')


@api.route(POST_MOVE_ENDPOINT, methods=['POST'])
def post_move():
    new_move = request.get_json()
    move = Move.query.get(new_move['id'])

    if move:
        return jsonify(result='success')

    if not move:
        move = Move.deserialize(new_move)

    if not move.valid:
        return jsonify(result='failed',
                       message=f"move {move.id} isn't valid."), 400

    db.session.add(move)
    try:
        db.session.commit()
    except IntegrityError:
        return jsonify(result='failed',
                       message="This node already has this move."), 400
    sent_node = Node()
    if 'sent_node' in new_move:
        sent_node.url = new_move['sent_node']

    move_broadcast.delay(
        move.id,
        sent_node_url=sent_node.url,
        my_node_url=f'{request.scheme}://{request.host}'
    )
    return jsonify(result='success')



from dataclasses import asdict
import json


class Logger:
    def __init__(self):
        self.logs = []
        self.print = False

    def log(self, log):
        self.logs.append(log)
        if self.print:
            print(log)

    def json_dump(self):
        status = []
        for log in self.logs:
            status.append(asdict(log))
        return json.dumps({'status': status})

    def get_characters(self):
        characters = {}
        for log in self.logs:
            if log.type == 'spawn':
                characters[log.id_] = {
                    'name': log.name,
                    'type': log.character_type,
                }
        return characters



from ..utils import itersubclasses


class ComponentContainer:
    def __init__(self):
        self.components = []

    def add_component(self, comp):
        comp.owner = self
        self.components.append(comp)
        return comp

    def remove_component(self, comp):
        comp.owner = None
        self.components.remove(comp)

    def get_component(self, cls):
        for comp in self.components:
            if comp.__class__ == cls:
                return comp
        subclasses = itersubclasses(cls)
        for subcls in subclasses:
            for comp in self.components:
                    if comp.__class__ == subcls:
                        return comp

    def get_components(self, cls):
        ret = []
        subclasses = itersubclasses(cls)
        for subcls in subclasses:
            for comp in self.components:
                    if comp.__class__ == subcls:
                        ret.append(comp)
        return ret


class Component:
    def __init__(self):
        self.owner = None

    def send_message(self, msg):
        for comp in self.owner.components:
            if not comp == self:
                comp.on_message(msg)

    def on_message(self, msg):
        # override me if need
        pass



import enum


class CharacterType(enum.IntEnum):
    NONE = 0
    PLAYER = 1
    MONSTER = 2


class ItemType(enum.IntEnum):
    ITEM = 0,
    FOOD = 1,
    HEAD = 2,
    ARMOR = 3,
    WEAPON = 4,


class AttackType(enum.IntEnum):
    MELEE = 0
    RANGED = 1
    MAGIC = 2



import datetime
import os
from typing import Mapping, Optional
import urllib.parse

from requests import Response, post
from requests.exceptions import ConnectionError, Timeout
from sqlalchemy.orm.query import Query

from .block import Block
from .node import Node
from .orm import db


__all__ = (
    'BROADCAST_LIMIT',
    'POST_BLOCK_ENDPOINT',
    'POST_MOVE_ENDPOINT',
    'POST_NODE_ENDPOINT',
    'broadcast_block',
    'broadcast_move',
    'broadcast_node',
)

BROADCAST_LIMIT = os.environ.get('BROADCAST_LIMIT', 100)
POST_BLOCK_ENDPOINT = '/blocks'
POST_MOVE_ENDPOINT = '/moves'
POST_NODE_ENDPOINT = '/nodes'


def broadcast_node(
        serialized: Mapping[str, str],
        sent_node: Optional[Node]=None,
        my_node: Optional[Node]=None,
) -> None:
    """
    It broadcast node to every nodes you know.

    :param      serialized: serialized :class:`nekoyume.node.Node`.
                            that will be broadcasted.
    :param       sent_node: sent :class:`nekoyume.node.Node`.
                            this node ignore sent node.
    :param         my_node: my :class:`nekoyume.node.Node`.
                            received node ignore my node when they
                            broadcast received object.
    """
    for node in filter_nodes(sent_node):
        try:
            url = urllib.parse.urljoin(node.url, POST_NODE_ENDPOINT)
            broadcast(url, serialized, node, my_node)
        except (ConnectionError, Timeout):
            continue

    db.session.commit()


def broadcast_block(
        serialized: Mapping[str, object],
        sent_node: Optional[Node]=None,
        my_node: Optional[Node]=None,
) -> None:
    """
    It broadcast this block to every nodes you know.

    :param      serialized: serialized :class:`nekoyume.block.Block`.
                            that will be broadcasted.
    :param       sent_node: sent :class:`nekoyume.node.Node`.
                            this node ignore sent node.
    :param         my_node: my :class:`nekoyume.node.Node`.
                            received node ignore my node when they
                            broadcast received object.
    """
    for node in filter_nodes(sent_node):
        try:
            url = urllib.parse.urljoin(node.url, POST_BLOCK_ENDPOINT)
            res = broadcast(url, serialized, node, my_node)
            if res.status_code == 403:
                result = res.json()
                # 0 is Genesis block.
                block_id = result.get('block_id', 0)
                query = db.session.query(Block).filter(
                    Block.id.between(block_id, serialized['id'])
                ).order_by(Block.id)
                offset = 0
                while True:
                    sync_blocks = query[
                        offset:offset+BROADCAST_LIMIT
                    ]
                    # TODO bulk api
                    for block in sync_blocks:
                        s = block.serialize(
                            use_bencode=False,
                            include_suffix=True,
                            include_moves=True,
                            include_hash=True
                        )
                        broadcast(url, s)
                    offset += BROADCAST_LIMIT
                    if len(sync_blocks) < BROADCAST_LIMIT:
                        break
        except (ConnectionError, Timeout):
            continue
    db.session.commit()


def broadcast_move(
        serialized: Mapping[str, object],
        sent_node: Optional[Node]=None,
        my_node: Optional[Node]=None,
) -> None:
    """
    It broadcast this move to every nodes you know.

    :param      serialized: serialized :class:`nekoyume.move.Move`.
                            that will be broadcasted.
    :param       sent_node: sent :class:`nekoyume.node.Node`.
                            this node ignore sent node.
    :param         my_node: my :class:`nekoyume.node.Node`.
                            received node ignore my node when they
                            broadcast received object.
    """
    for node in filter_nodes(sent_node):
        try:
            url = urllib.parse.urljoin(node.url, POST_MOVE_ENDPOINT)
            broadcast(url, serialized, node, my_node)
        except (ConnectionError, Timeout):
            continue

    db.session.commit()


def filter_nodes(sent_node: Optional[Node]=None) -> Query:
    query = db.session.query(Node)
    if sent_node:
        query = query.filter(
            Node.url != sent_node.url
        )
    return query


def broadcast(url: str, serialized: Mapping[str, object],
              node: Optional[Node]=None,
              sent_node: Optional[Node]=None) -> Response:
    if sent_node:
        serialized['sent_node'] = sent_node.url
    res = post(url, json=serialized, timeout=3)
    if node:
        node.last_connected_at = datetime.datetime.utcnow()
        db.session.add(node)
    return res



import enum

from .. import Component


# for behavior components
class Behavior(Component):
    pass


class BehaviorTreeBuilder:
    def __init__(self):
        self.node = None
        self.parents = []

    def do(self, name, fn):
        if not self.parents:
            raise RuntimeError(
                "Can't create an unnested ActionNode, it must be a leaf node"
            )
        self.parents[-1].add_child(ActionNode(name, fn))
        return self

    def condition(self, name, fn):
        BTS = BehaviorTreeStatus
        return self.do(name, lambda b: BTS.SUCCESS if fn(b) else BTS.FAILURE)

    def sequence(self, name):
        node = SequenceNode(name)
        if self.parents:
            self.parents[-1].add_child(node)
        self.parents.append(node)
        return self

    def selector(self, name):
        node = SelectorNode(name)
        if self.parents:
            self.parents[-1].add_child(node)
        self.parents.append(node)
        return self

    def end(self):
        self.node = self.parents.pop()
        return self

    def build(self):
        if not self.node:
            raise RuntimeError(
                "Zero node."
            )
        return self.node


class BehaviorTreeStatus(enum.Enum):
    SUCCESS = 1,
    FAILURE = 2,
    RUNNING = 3,


class BehaviorNode:
    def __init__(self, name):
        self.name = name


class ActionNode(BehaviorNode):
    def __init__(self, name, fn):
        super().__init__(name)
        self.fn = fn

    def tick(self, data):
        return self.fn(data)


class SequenceNode(BehaviorNode):
    def __init__(self, name):
        super().__init__(name)
        self.children = []

    def add_child(self, node):
        self.children.append(node)

    def tick(self, data):
        for child in self.children:
            status = child.tick(data)
            if status != BehaviorTreeStatus.SUCCESS:
                return status
        return BehaviorTreeStatus.SUCCESS


class SelectorNode(BehaviorNode):
    def __init__(self, name):
        super().__init__(name)
        self.children = []

    def add_child(self, node):
        self.children.append(node)

    def tick(self, data):
        for child in self.children:
            status = child.tick(data)
            if status != BehaviorTreeStatus.FAILURE:
                return status
        return BehaviorTreeStatus.FAILURE



import os
import time

from babel.messages.frontend import compile_catalog
from click import IntRange, ParamType, argument, echo, group, option
from coincurve import PrivateKey
from ptpython.repl import embed

from .app import app
from .block import Block
from .broadcast import broadcast_block, broadcast_node
from .move import Move, get_my_public_url
from .node import Node
from .orm import db
from .user import User


DEFAULT_SEED_NODE_URL = os.environ.get(
    'SEED_NODE_URL',
    'http://seed.nekoyu.me'
)


class PrivateKeyType(ParamType):
    name = 'private key'

    def convert(self, value, param, ctx) -> PrivateKey:
        val = value[2:] if value.startswith(('0x', '0X')) else value
        try:
            num = bytes.fromhex(val)
            return PrivateKey(num)
        except (ValueError, TypeError):
            self.fail('%s is not a valid private key of 64 hexadecimal digits')


@group()
def cli():
    pass


@cli.command()
@argument('private_key', type=PrivateKeyType())
@option('--sleep', default=0, type=float)
def mine(private_key: PrivateKey, sleep: float):
    app.app_context().push()

    while True:
        Block.sync()
        block = Block.create(
            User(private_key),
            [m
             for m in Move.query.filter_by(block=None).limit(20).all()
             if m.valid],
            echo=echo,
            sleep=sleep,
        )
        if block:
            serialized = block.serialize(
                use_bencode=False,
                include_suffix=True,
                include_moves=True,
                include_hash=True
            )
            broadcast_block(serialized=serialized)
            echo(block)


@cli.command()
def shell():
    app.app_context().push()
    embed(globals(), locals())


@cli.command()
@option('--seed',
        default=DEFAULT_SEED_NODE_URL,
        help='Seed node URL to connect')
@option('--sync/--skip-sync',
        default=False,
        help='Synchronize after initialization or skip it')
def init(seed, sync):
    echo('Creating database...')
    db.create_all()
    echo(f'Updating node... (seed: {seed})')
    if sync:
        Node.update(Node.get(url=seed))
        echo('Syncing blocks...')
        Block.sync(echo=echo)

    echo('Compiling translations...')
    dir_path = os.path.abspath(os.path.dirname(__file__))
    compile_command = compile_catalog()
    compile_command.directory = dir_path + '/translations'
    compile_command.finalize_options()
    compile_command.run()


@cli.command()
@option('--seed',
        default=DEFAULT_SEED_NODE_URL,
        help='Seed node URL to connect')
def sync(seed: str):
    public_url = get_my_public_url()
    if public_url:
        echo(f"You have a public node url. ({public_url})")
        broadcast_node(serialized={'url': public_url})
    Node.update(Node.get(url=seed))
    engine = db.engine
    if not engine.dialect.has_table(engine.connect(), Block.__tablename__):
        echo("You need to initialize. try `nekoyume init`.")
        return False
    while True:
        try:
            prev_id = Block.query.order_by(Block.id.desc()).first().id
        except AttributeError:
            prev_id = 0
        Block.sync(echo=echo)
        try:
            if prev_id == Block.query.order_by(Block.id.desc()).first().id:
                echo("The blockchain is up to date.")
                time.sleep(15)
        except AttributeError:
            echo(("There is no well-connected node. "
                  "please check you network."))
            break


@cli.command()
def doctor():
    id = 1
    for block in Block.query.order_by(Block.id.asc()):
        if block.id != id:
            echo(f'Block {id}: is empty.')
            id = block.id
        if not block.valid:
            echo(f'Block {id}: is invalid.')
        id += 1


@cli.command()
def repair():
    id = 1
    for block in Block.query.order_by(Block.id.asc()):
        if block.id != id or not block.valid:
            Block.query.filter(Block.id >= block.id).delete(
                synchronize_session='fetch'
            )
            echo(f'Block {id}+ was removed.')
            break
        id += 1

    deleted_move_ids = []
    for move in Move.query:
        if not move.valid:
            move.delete()
            deleted_move_ids.append(move.id)

    if deleted_move_ids:
        echo(f'Following moves were removed.')
        for deleted_move_id in deleted_move_ids:
            echo(f'   {deleted_move_id}')

    db.session.commit()


@cli.command()
@option('--host',
        default='127.0.0.1',
        help='Host to listen')
@option('--port',
        type=IntRange(0, 65535),
        metavar='PORT',
        default=5000,
        help='Port number to listen')
def dev(host: str, port: int) -> None:
    app.run(debug=True, host=host, port=port)


if __name__ == '__main__':
    cli()



import codecs
import os

from setuptools import find_packages, setup


here = os.path.abspath(os.path.dirname(__file__))


def long_description():
    try:
        with codecs.open(os.path.join(here, 'README.rst'),
                         encoding='utf-8') as f:
            return f.read()
    except IOError:
        pass


install_requires = [
    'bencode.py >= 2.0.0, < 2.1.0',
    'blinker >= 1.4, < 1.5',
    'celery >= 4.1.0, < 4.2.0',
    'click >= 6.7, < 7.0',
    'cykeccak >= 0.13.2.2, < 0.14.0.0',
    'dataclasses >= 0.6, < 1.0',
    'Flask >= 0.12.2, < 0.13.0',
    'Flask-Babel >= 0.11.2, < 0.12.0',
    'Flask-Caching >= 1.4.0, < 1.5.0',
    'Flask-SQLAlchemy >= 2.3.2, < 2.4.0',
    'gevent == 1.3.5',
    'gunicorn >= 19.9.0',
    'psycopg2 >= 2.7.5, < 2.8.0',
    'ptpython == 0.41',
    'pytz >= 2018.3',
    'sentry-sdk[flask] == 0.4.1',
    'redis >= 2.10.6, < 2.11.0',
    'requests >= 2.18.4, < 2.19.0',
    'coincurve >= 9.0.0',
    'SQLAlchemy >= 1.2.2, < 1.3.0',
    'tablib >= 0.12.1, < 0.13.0',
    'typeguard >= 2.2.2, < 2.3.0',
]

if __name__ == '__main__':
    setup(
        name='nekoyume',
        version='0.1.3',
        description='Decentralized MMORPG based on Dungeon World',
        long_description=long_description(),
        url='https://github.com/nekoyume/nekoyume',
        license='LICENSE.txt',
        author='JC Kim',
        author_email='jc@nekoyu.me',
        classifiers=[
            'Development Status :: 2 - Pre-Alpha',
            'Intended Audience :: Other Audience',
            'Topic :: Database :: Database Engines/Servers',
            'Topic :: Games/Entertainment :: Role-Playing',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3.6',
        ],
        keywords='blockchain mmorpg game',
        packages=find_packages(exclude=['contrib', 'docs', 'tests']),
        install_requires=install_requires,
        extras_require={
            'dev': [
                'flake8 >= 3.5.0, < 3.6.0',
                'flake8-import-order-spoqa >= 1.5.0, < 1.6.0',
                'recommonmark==0.4.0',
                'Sphinx >= 1.7.1, < 1.8.0',
                'sphinx-rtd-theme==0.2.4',
            ],
            'test': [
                'codecov >= 2.0.15, < 2.1.0',
                'pytest >= 3.8.0, < 3.9.0',
                'pytest-flake8 >= 1.0.2, < 1.1.0',
                'pytest-localserver >= 0.4.1, < 0.5',
                'requests-mock >= 1.5.2, < 1.6.0',
            ],
        },
        package_data={
            'nekoyume': [
                'tables/*.tsv'
                'templates/*.html',
                'translations/*/LC_MESSAGES/*'
                ],
        },
        entry_points={
            'console_scripts': [
                'nekoyume = nekoyume.cli:cli',
            ],
        },
        project_urls={
            'Bug Reports': 'https://github.com/nekoyume/nekoyume/issues',
            'Funding': 'https://nekoyu.me/',
            'Source': 'https://github.com/nekoyume/nekoyume/',
        },
    )



from ..battle.enums import AttackType, ItemType
from .base import Equipment


class Weapon(Equipment):
    def __init__(self, name):
        super().__init__(name)
        self.type_ = ItemType.WEAPON
        self.atk_type = AttackType.MELEE

    def get_atk(self):
        return self.data.param_0


class RangedWeapon(Weapon):
    def __init__(self, name):
        super().__init__(name)
        self.atk_type = AttackType.RANGED



class InvalidNameError(Exception):
    pass


class InvalidMoveError(Exception):
    pass


class InvalidBlockError(Exception):
    pass


class NodeUnavailable(Exception):
    pass


class OutOfRandomError(Exception):
    pass



from . import Behavior, BehaviorTreeStatus


class Aggro(Behavior):
    def __init__(self):
        self.value = 1  # my global aggro
        self.targets = {}  # target aggro

    def tick(self, simulator):
        self.value = max(1, self.value - 1)
        for v in self.targets:
            self.targets[v] = max(0, self.targets[v] - 1)
        return BehaviorTreeStatus.SUCCESS

    def add(self, target_id, value):
        if target_id in self.targets:
            self.targets[target_id] += value
        else:
            self.targets[target_id] = value



from celery import Celery

from .block import Block
from .broadcast import broadcast_block, broadcast_move
from .move import Move
from .node import Node
from .orm import db


celery = Celery()


@celery.task()
def move_broadcast(move_id, sent_node_url, my_node_url, session=db.session):
    move = session.query(Move).get(move_id)
    if not move:
        return
    serialized = move.serialize(
        use_bencode=False,
        include_signature=True,
        include_id=True,
    )
    broadcast_move(
        serialized=serialized,
        sent_node=Node(url=sent_node_url),
        my_node=Node(url=my_node_url)
    )


@celery.task()
def block_broadcast(block_id, sent_node_url, my_node_url, session=db.session):
    block = session.query(Block).get(block_id)
    if not block:
        return
    serialized = block.serialize(
        use_bencode=False,
        include_suffix=True,
        include_moves=True,
        include_hash=True
    )
    broadcast_block(
        serialized=serialized,
        sent_node=Node(url=sent_node_url),
        my_node=Node(url=my_node_url)
    )



class WeightedList:
    def __init__(self):
        self.values_ = []
        self.weights_ = []

    def __len__(self):
        return len(self.values_)

    def __str__(self):
        return str((self.values_, self.weights_))

    def add(self, value, weight):
        self.values_.append(value)
        self.weights_.append(weight)

    def select(self, random, pop=False):
        if not self.values_:
            return None
        weight_sum = 0
        for i in self.weights_:
            weight_sum += i
        rnd = random.randint(0, weight_sum - 1)
        idx = -1
        for i in range(len(self.values_)):
            if rnd < self.weights_[i]:
                idx = i
                break
            rnd -= self.weights_[i]
        if idx < 0:
            return None
        ret = self.values_[idx]
        if pop:
            del self.values_[i]
            del self.weights_[i]
        return ret

    def pop(self, random):
        return self.select(random, True)



import functools

from coincurve import PrivateKey
from flask import (Blueprint, Response, g, redirect, render_template, request,
                   session, url_for)
from flask_babel import Babel
from sqlalchemy import func

from .broadcast import broadcast_move
from .move import LevelUp, Move
from .node import Node
from .orm import db
from .user import User, cache


game = Blueprint('game', __name__, template_folder='templates')
babel = Babel()


@babel.localeselector
def get_locale():
    langs = {babel.default_locale}
    langs.update(babel.list_translations())
    return request.accept_languages.best_match([l.language for l in langs])


def login_required(f):
    @functools.wraps(f)
    def decorated_function(*args, **kwargs):
        private_key_hex = session.get('private_key')
        error = None
        if private_key_hex is not None:
            if private_key_hex.startswith(('0x', '0X')):
                private_key_hex = private_key_hex[2:]
            try:
                private_key_bytes = bytes.fromhex(private_key_hex)
                private_key = PrivateKey(private_key_bytes)
            except (ValueError, TypeError):
                error = 'invalid-private-key'
            else:
                g.user = User(private_key)
                return f(*args, **kwargs)
        return redirect(url_for('.get_login', next=request.url, error=error))
    return decorated_function


@game.route('/login', methods=['GET'])
def get_login():
    return render_template('login.html')


@game.route('/login', methods=['POST'])
def post_login():
    session['name'] = request.values.get('name')
    session['private_key'] = request.values.get('private_key')
    if 'next' in request.values:
        return redirect(request.values.get('next'))
    else:
        return redirect(url_for('.get_game'))


@game.route('/logout', methods=['GET'])
@login_required
def get_logout():
    del session['private_key']
    return redirect(url_for('game.get_login'))


@cache.memoize(60)
def get_rank():
    return db.session.query(
        LevelUp.user_address, func.count(LevelUp.id)
    ).group_by(LevelUp.user_address).order_by(
        func.count(LevelUp.id).desc()
    ).limit(10).all()


def get_unconfirmed_move(address):
    unconfirmed_moves = Move.query.filter_by(
        user_address=address, block=None
    )
    for unconfirmed_move in unconfirmed_moves:
        if unconfirmed_move.valid:
            return unconfirmed_move
    return None


@game.route('/dashboard')
@login_required
def get_dashboard():
    if not g.user.avatar():
        return redirect(url_for('.get_new_novice'))

    unconfirmed_move = get_unconfirmed_move(g.user.address)

    feed = g.user.moves
    # for caching
    for move in reversed(feed.limit(10).all()):
        avatar, result = move.execute()
    return render_template('dashboard.html',
                           unconfirmed_move=unconfirmed_move,
                           feed=feed.order_by(Move.block_id.desc()),
                           rank=get_rank())


@game.route('/')
@login_required
def get_game():
    avatar = g.user.avatar()
    if not avatar:
        return redirect(url_for('.get_new_novice'))
    if avatar.class_ == 'novice':
        return redirect(url_for('.get_first_class'))

    return render_template('game.html')


@game.route('/status')
@login_required
def get_status():
    if not g.user.avatar():
        return redirect(url_for('.get_new_novice'))

    feed = g.user.moves
    # for caching
    for move in reversed(feed.limit(10).all()):
        avatar, result = move.execute()
    return render_template('status.html',
                           feed=feed.order_by(Move.block_id.desc()))


@game.route('/in_progress')
@login_required
def get_unconfirmed():
    if not g.user.avatar():
        return redirect(url_for('.get_new_novice'))

    unconfirmed_move = get_unconfirmed_move(g.user.address)
    if unconfirmed_move:
        return "true"
    return "false"


@game.route('/new')
@login_required
def get_new_novice():
    if not g.user.avatar():
        move = Move.query.filter_by(
            user_address=g.user.address,
            name='create_novice',
        ).first()
        if not move:
            move = g.user.create_novice({'name': session.get('name', '')})
            db.session.add(move)
            db.session.commit()
            serialized = move.serialize(
                use_bencode=False,
                include_signature=True,
                include_id=True,
            )
            broadcast_move(
                serialized,
                my_node=Node(url=f'{request.scheme}://{request.host}')
            )
        return render_template('new.html', move=move)
    return redirect(url_for('.get_game'))


@game.route('/first_class')
@login_required
def get_first_class():
    avatar = g.user.avatar()
    if not avatar:
        return redirect(url_for('.get_new_novice'))
    if avatar.class_ != 'novice':
        return redirect(url_for('.get_game'))

    return render_template('first_class.html')


@game.route('/session_moves', methods=['POST'])
@login_required
def post_move():
    unconfirmed_move = get_unconfirmed_move(g.user.address)

    if unconfirmed_move:
        return redirect(url_for('.get_game'))

    if request.values.get('name') == 'hack_and_slash':
        if g.user.avatar().dead:
            return redirect(url_for('.get_game'))
        move = g.user.hack_and_slash(request.values.get('weapon'),
                                     request.values.get('armor'),
                                     request.values.get('food'),)
    if request.values.get('name') == 'sleep':
        move = g.user.sleep()
    if request.values.get('name') == 'level_up':
        move = g.user.level_up(request.values.get('new_status'))
    if request.values.get('name') == 'say':
        move = g.user.say(request.values.get('content'))
    if request.values.get('name') == 'send':
        move = g.user.send(request.values.get('item'),
                           request.values.get('amount'),
                           request.values.get('receiver'))
    if request.values.get('name') == 'combine':
        move = g.user.combine(request.values.get('item1'),
                              request.values.get('item2'),
                              request.values.get('item3'))
    if request.values.get('name') == 'first_class':
        move = g.user.first_class(request.values.get('class_'))
    if request.values.get('name') == 'move_zone':
        move = g.user.move_zone(request.values.get('zone'))

    if move:
        serialized = move.serialize(
            use_bencode=False,
            include_signature=True,
            include_id=True,
        )
        broadcast_move(
            serialized,
            my_node=Node(url=f'{request.scheme}://{request.host}')
        )
    return redirect(url_for('.get_game'))


@game.route('/export/', methods=['GET'])
@login_required
def export_private_key():
    key = session['private_key']
    file_name = f'{g.user.address}.csv'
    return Response(
        key,
        headers={
            'Content-Disposition': f'attachment;filename={file_name}',
            'Content-Type': 'text/csv',
        }
    )



import datetime

from requests import get
from requests.exceptions import ConnectionError, Timeout
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.session import Session

from .orm import db


class Node(db.Model):
    """This object contains node information you know."""

    #: URL of node
    url = db.Column(db.String, primary_key=True)
    #: last connected datetime of the node
    last_connected_at = db.Column(db.DateTime, nullable=False, index=True)

    get_nodes_endpoint = '/nodes'
    get_blocks_endpoint = '/blocks'

    @classmethod
    def get(cls, url: str, session: Session=db.session):
        get_node = Node.query.filter_by(url=url).first
        node = get_node()
        if node:
            return node
        elif get(f'{url}/ping').text == 'pong':
            node = Node(url=url, last_connected_at=datetime.datetime.utcnow())
            if session:
                session.add(node)
                try:
                    session.commit()
                except IntegrityError:
                    node = get_node()
                    if node is None:
                        raise
                    return node
            return node
        else:
            return None

    @classmethod
    def update(cls, node: 'Node'):
        """
        Update recent node list by scrapping other nodes' information.
        """
        try:
            response = get(f"{node.url}{Node.get_nodes_endpoint}")
        except (ConnectionError, Timeout):
            return
        for url in response.json()['nodes']:
            try:
                Node.get(url)
            except (ConnectionError, Timeout):
                continue
        db.session.commit()

    def ping(self):
        try:
            result = get(f'{self.url}/ping').text == 'pong'
            if result:
                self.last_connected_at = datetime.datetime.utcnow()
            return result
        except (ConnectionError, Timeout):
            return False
